#include "stdlib.h"
#include "stdio.h"
#include "math.h"
#include "string.h"

#include "WavSigGen.h"

// Needed for output status
double sineInDB;
double modRangeInPercent;

// Here we store the app information
FileObject wavFile;
PCMParameters pcm;
LFOParameters lfo;
LFOParameters phaseModulator;

#ifdef SHOW_STATUS
	char status;
#endif

/**
 * Returns if
 * value = 20 * log10(x)
 */
double dBToLinear(double value)
{
	return pow(10.0, value / 20.0);
}

/**
 * Returns modulator range
 * @param percent
 * Value from 0 to 100% where 100% equals 2*PI
 */
double getModRange(double percent) {
	return TWO_PI * (percent / 100.0);
}

/**
 * Parses the command line arguments and sets
 * the local values according to the user input
 */
int parseArguments(int numArgs, const char* arg[]) 
{
	
	// Parse arguments
	for (int i = 0; i < numArgs - 1; i++) 
	{	
		// Get the value
		const char* value = arg[i + 1];
		
		//
		// DROP HELP MESSAGE
		//
		if (strcmp(arg[i], HELP_FLAG) == 0 || strcmp(value, HELP_FLAG) == 0) 
		{
			return ERROR;
		}		
		else
		//
		// OUTPUT FILENAME
		//
		if (strcmp(arg[i], FILENAME_FLAG) == 0) 
		{
			if ((int)strlen(value) == 0) 
			{
				printf("ERROR: Illegal filename or filename empty.\n");
				return ERROR;
			}
			wavFile.name = value;
		}		
		else
		//
		// OUTPUT FILESIZE
		//
		if (strcmp(arg[i], FILESIZE_FLAG) == 0) 
		{
			int tmp = atoi(value);
			if (tmp < MIN_FILESIZE_IN_KB || tmp > MAX_FILESIZE_IN_KB) 
			{
				printf("ERROR: Illegal filesize. Enter a value between %i and %i kB.\n", MIN_FILESIZE_IN_KB, MAX_FILESIZE_IN_KB);
				return ERROR;
			}
			
			wavFile.numBytes = tmp * NUM_BYTES_PER_KB;
		}
		else
		//
		// NUM CHANNELS
		//
		if (strcmp(arg[i], NUM_CHANNELS_FLAG) == 0) 
		{
			int tmp = atoi(value);
			if (tmp < MIN_NUM_CHANNELS || tmp > MAX_NUM_CHANNELS) 
			{
				printf("ERROR: Illegal number of channels. Enter a value between %i and %i.\n", MIN_NUM_CHANNELS, MAX_NUM_CHANNELS);
				return ERROR;
			}

			pcm.numChannels = tmp;
		}
		else
		//
		// NUM BYTES PER SAMPLE
		//
		if (strcmp(arg[i], SAMPLESIZE_FLAG) == 0) 
		{
			int tmp = atoi(value);
			if (tmp < MIN_NUM_BYTES_PER_SAMPLE || tmp > MAX_NUM_BYTES_PER_SAMPLE) 
			{
				printf("ERROR: Illegal samplesize. Enter a value between %i and %i bytes.\n", MIN_NUM_BYTES_PER_SAMPLE, MAX_NUM_BYTES_PER_SAMPLE);
				return ERROR;
			}

			pcm.numBytesPerSample = tmp;
		}
		else
		//
		// SAMPLERATE
		//
		if (strcmp(arg[i], SAMPLERATE_FLAG) == 0) 
		{
			int tmp = atoi(value);
			if (tmp < MIN_SAMPLERATE_HZ || tmp > MAX_SAMPLERATE_HZ) 
			{
				// Drop an errormsg
				printf("Illegal samplerate. Enter a value between %if and %if Hz.\n", MIN_SAMPLERATE_HZ, MAX_SAMPLERATE_HZ);
				return ERROR;
			}

			pcm.sampleRateHz = tmp;
		}
		else
		//
		// SINE FREQUENCY
		//
		if (strcmp(arg[i], SINE_FREQ_FLAG) == 0) 
		{
			double tmp = atof(value);
			if (tmp < MIN_SINE_FREQ_HZ || tmp > MAX_SINE_FREQ_HZ) 
			{
				// Drop an errormsg
				printf("Illegal frequency. Enter a value between %.2f Hz and %.2f Hz.\n", MIN_SINE_FREQ_HZ, MAX_SINE_FREQ_HZ);
				return ERROR;
			}

			lfo.freqHz = tmp;
		}
		//
		// AMPLITUDE IN DB
		//
		if (strcmp(arg[i], DAMPING_FLAG) == 0) 
		{
			double tmp = atof(value);
			if (tmp < MIN_DAMPING_DB || tmp > MAX_DAMPING_DB) 
			{
				// Drop an errormsg
				printf("Illegal damping. Enter a value between %.2f and %.2f dB.\n", MIN_DAMPING_DB, MAX_DAMPING_DB);
				return ERROR;
			}
			
			// Calculate linear value
			lfo.amplitude = dBToLinear(tmp);
			sineInDB = tmp;
		}
		//
		// PHASE MOD FREQ IN HZ
		//
		if (strcmp(arg[i], MODULATE_FREQ_FLAG) == 0) 
		{
			double tmp = atof(value);
			if (tmp < MIN_MOD_FREQ_HZ || tmp > MAX_MOD_FREQ_HZ) 
			{
				// Drop an errormsg
				printf("Illegal modulator frequency. Enter a value between %.2f and %.2f Hz.\n", MIN_MOD_FREQ_HZ, MAX_MOD_FREQ_HZ);
				return ERROR;
			}

			phaseModulator.freqHz = tmp;
		}
		//
		// PHASE MOD RANGE IN %
		//
		if (strcmp(arg[i], MODULATE_RANGE_FLAG) == 0) 
		{
			double tmp = atof(value);
			if (tmp < MIN_MOD_RANGE_PERCENT || tmp > MAX_MOD_RANGE_PERCENT) 
			{
				// Drop an errormsg
				printf("Illegal modulator phase range. Enter a value between %.2f and %.2f percent.\n", MIN_MOD_RANGE_PERCENT, MAX_MOD_RANGE_PERCENT);
				return ERROR;
			}
			
			// Calculate phase range
			phaseModulator.amplitude = getModRange(tmp);
			modRangeInPercent = tmp;
		}
	}

	return VALID;
}

/**
 * Writes the bytes of valut to the output stream
 * @param numBytes
 * The number of bytes which shall be written
 * @param value
 * The value that contains the bytes that shall be written
 */
void writeLittleEndian(FILE* file, unsigned int value, int numBytes) 
{	
	// Sanity check
	if (numBytes <= sizeof(int)) {
		int numBits = (numBytes * SIZEOF_BYTE);
		for (int i = 0; i < numBits; i += 8) 
		{
			fputc(((value >> i) & 0xFF), file);
		}
	}
}

/**
 * Writes the char array to the file output stream.
 */
void writeLittleEndian(FILE* file, const char* str) 
{
	int len = (int)strlen(str);
	for (int i = 0; i < len; i++) 
	{
		fputc(str[i], file);
	}
}

/**
 * Writes the wave header to the file output stream.
 * This function relies on the PCMParameters of the pcm struct
 * to work correctly. Hence, those values have to be set before
 * calling this function.
 * 
 * @return
 * The number of PCM frames that have
 * been defined to be written.
 */
int writeWAVheader(FILE* file) 
{
	int numBytesPerFrame = pcm.numBytesPerSample * pcm.numChannels;
	int numBytesPerSecond = pcm.sampleRateHz * numBytesPerFrame;
	
	// Calculate proper chunksize
	int numDataBytes = numBytesPerFrame * (int)ceil(wavFile.numBytes / (double) numBytesPerFrame);

	writeLittleEndian(file, RIFF_STRING);
	writeLittleEndian(file, NUM_BYTES_WAVE_FILE_HEADER + numDataBytes - 8, 4);
	writeLittleEndian(file, WAVE_STRING);
	writeLittleEndian(file, NUM_BYTES_FMT_HEADER, 4);
	writeLittleEndian(file, PCM_FORMAT, 2);
	writeLittleEndian(file, pcm.numChannels, 2);
	writeLittleEndian(file, pcm.sampleRateHz, 4);
	writeLittleEndian(file, numBytesPerSecond, 4);
	writeLittleEndian(file, numBytesPerFrame, 2);
	writeLittleEndian(file, pcm.numBytesPerSample * SIZEOF_BYTE, 2);
	writeLittleEndian(file, DATA_STRING);
	writeLittleEndian(file, numDataBytes, 4);
	
	// Return the number of frames we're going to write
	return (numDataBytes / numBytesPerFrame);
}

/**
 * Writes the wave file including the wave header by calling the
 * writeWAVHeader function.
 * The wave file contains the sound of a sine oscillator (LFO)
 * which has been configured by the user arguments.
 * 
 * @return
 * VALID wave file was sucessfully written
 * ERROR error opening the wave file
 */
int writeWAVFile() 
{	
	if (wavFile.name == 0) 
	{
		return ERROR;
	}

	// Open file in binary mode
	FILE* fp = fopen(wavFile.name, "wb");
	if (fp == NULL) 
	{
		return ERROR;
	}

	// Write header first
	int numFrames = writeWAVheader(fp);

	int numBitsPerSample = pcm.numBytesPerSample * SIZEOF_BYTE;
	
	// Multily linear amplitude value with peak value
	lfo.amplitude *= (1 << (numBitsPerSample - 1)) - 1;
	
	double secondPerRad = (TWO_PI / (double)pcm.sampleRateHz);
	
	// Increment values
	double radPerSample = lfo.freqHz * secondPerRad;
	double radPerSampleMod = phaseModulator.freqHz * secondPerRad;
	
	lfo.phase = 0.0;
	phaseModulator.phase = 0.0;

	// Loop over the number of frames
	for (int i = numFrames; --i >= 0; ) 
	{
		double phaseUsed = lfo.phase;
		
		// Check phase modulator
		if (phaseModulator.freqHz != 0)
		{
			// Modulate phase
			phaseUsed *= phaseModulator.amplitude * sin(phaseModulator.phase);
			phaseModulator.phase += radPerSampleMod;
		
			if (phaseModulator.phase >= TWO_PI) {
				phaseModulator.phase -= TWO_PI;
			}
		}

		// Create integer sample
		int sample = (int)(lfo.amplitude * sin(phaseUsed));
		lfo.phase += radPerSample;
		
		if (lfo.phase >= TWO_PI) {
			lfo.phase -= TWO_PI;
		}

		// Every channel has the same content
		for (int c = pcm.numChannels; --c >= 0; ) 
		{
			for (int k = 0; k < numBitsPerSample; k += SIZEOF_BYTE) 
			{
				fputc((sample >> k) & 0xFF, fp);
			}
		}

#ifdef SHOW_STATUS
		if (i % 100000 == 0) {
			status = (status == '-') ? '|' : '-';
			printf("%c%c", (char)(0x8), status);
		}
#endif

	}

	fclose(fp);
	return VALID;
}

/**
 * Drops an argument help message 
 * on the command line
 */
void dropHelpMessage() 
{
	printf("\n");
	printf("Wave generator options:\n");
	printf("  %s\t show this option menu\n", HELP_FLAG);
	printf("File:\n");
	printf("  %s\t the name of the output file (default: %s)\n", FILENAME_FLAG, DEFAULT_FILENAME);
	printf("  %s\t output filesize in kB (default: %i)\n", FILESIZE_FLAG, DEFAULT_FILESIZE_IN_KB );
	printf("  %s\t the sample rate in Hz (default: %i)\n", SAMPLERATE_FLAG, DEFAULT_SAMPLERATE_HZ);
	printf("  %s\t the number of channels (default: %i)\n", NUM_CHANNELS_FLAG, DEFAULT_NUM_CHANNELS);
	printf("  %s\t the number of bytes per sample (default: %i)\n", SAMPLESIZE_FLAG, DEFAULT_NUM_BYTES_PER_SAMPLE);
	printf("Sine  LFO:\n");
	printf("  %s\t LFO frequency in Hz (default: %.2f)\n", SINE_FREQ_FLAG, DEFAULT_SINE_FREQ_HZ);
	printf("  %s\t attenuation in dB (default: %.2f)\n", DAMPING_FLAG, DEFAULT_DAMPING_DB);
	printf("Phase LFO:\n");
	printf("  %s\t modulation frequency in Hz (default: %.2f)\n", MODULATE_FREQ_FLAG, DEFAULT_MOD_FREQ_HZ);
	printf("  %s\t modulation range in percent (100%% = 2PI) (default: %.2f)\n", MODULATE_RANGE_FLAG, DEFAULT_MOD_RANGE_PERCENT);
}

/**
 * MAIN
 */
int main(int numArgs, const char* arg[])
{	
	// Init PCM parameters, LFO and output file
	// parameters with default values
	pcm.numBytesPerSample = DEFAULT_NUM_BYTES_PER_SAMPLE;
	pcm.sampleRateHz = DEFAULT_SAMPLERATE_HZ;
	pcm.numChannels = DEFAULT_NUM_CHANNELS;
	wavFile.name = DEFAULT_FILENAME;
	wavFile.numBytes = DEFAULT_FILESIZE_IN_KB * NUM_BYTES_PER_KB;
	lfo.freqHz = DEFAULT_SINE_FREQ_HZ;
	lfo.amplitude = dBToLinear(DEFAULT_DAMPING_DB);
	phaseModulator.freqHz = DEFAULT_MOD_FREQ_HZ;
	phaseModulator.amplitude = getModRange(DEFAULT_MOD_RANGE_PERCENT);
	sineInDB = DEFAULT_DAMPING_DB;
	modRangeInPercent = DEFAULT_MOD_RANGE_PERCENT;

	// Parse the arguments
	int retCode = parseArguments(numArgs, arg);
	if (retCode == VALID) 
	{	
		// Drop status MSG
		printf("Generating %s:\n", wavFile.name);
		printf("Sine  LFO : %.2f Hz, max  : %.2f dB\n", lfo.freqHz, sineInDB);
		printf("Phase LFO : %.2f Hz, range: %.2f percent\n", phaseModulator.freqHz, modRangeInPercent);
		printf("PCM       : %i Bit, %i Hz, %i Ch\n", (pcm.numBytesPerSample * 8), pcm.sampleRateHz, pcm.numChannels);	
		
		// Write the wave file
		retCode = writeWAVFile();
		if (retCode == VALID)
		{
			printf("done\n");
		}
		else
		{
			printf("Error writing wave file!");
		}
	}    
	else 
	{
		dropHelpMessage();
	}

	return retCode;
}

